__all__ = [
    "SEARCH_OP_EQ",
    "SEARCH_OP_LT",
    "SEARCH_OP_LE",
    "SEARCH_OP_GT",
    "SEARCH_OP_GE",

    "SEARCH_OP_CO",
    "SEARCH_OP_ICO",

    "SEARCH_OP_IN",

    "SEARCH_OP_ISW",
    "SEARCH_OP_IEW",
    "SEARCH_OP_LIKE",
    "SEARCH_OP_ILIKE",

    "SEARCH_OPERATIONS",
]


SEARCH_OP_EQ = "eq"
SEARCH_OP_LT = "lt"
SEARCH_OP_LE = "le"
SEARCH_OP_GT = "gt"
SEARCH_OP_GE = "ge"

SEARCH_OP_CO = "co"
SEARCH_OP_ICO = "ico"

SEARCH_OP_IN = "in"

SEARCH_OP_ISW = "isw"
SEARCH_OP_IEW = "iew"
SEARCH_OP_LIKE = "like"
SEARCH_OP_ILIKE = "ilike"

SEARCH_OPERATIONS = (
    SEARCH_OP_EQ,
    SEARCH_OP_LT,
    SEARCH_OP_LE,
    SEARCH_OP_GT,
    SEARCH_OP_GE,
    SEARCH_OP_IN,
    SEARCH_OP_CO,
    SEARCH_OP_ICO,
    SEARCH_OP_ISW,
    SEARCH_OP_IEW,
    SEARCH_OP_LIKE,
    SEARCH_OP_ILIKE,
    SEARCH_OP_ILIKE,
)
